#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>

#define DOTS 5000

#define WIN_W		256
#define WIN_H		256
#define WIN_X		0
#define WIN_Y		0

typedef struct {
    int vweight;
    int hweight;
    GC gc;
} SAND;

typedef struct {
    short x,y;
    SAND sand;
    int dirn;
    int dirs;
    int dire;
    int dirw;
} DOT;

SAND *pos, *opos, *tpos;
SAND pos1[WIN_W*WIN_H];
SAND pos2[WIN_W*WIN_H];
DOT dots[DOTS*2];
XPoint points[DOTS*2];

char		*progname;
Display		*display;
int		screen;
Colormap	cmap;
Window		window;
unsigned long	pen_fg, pen_bg;
XEvent		event;    
GC		gcblack, gcs[7];
Pixmap		background;

void InitWindow(int argc, char *argv[])
{
    XSizeHints	size_hints;
    XWMHints	wm_hints;
    XClassHint	class_hints;
    XGCValues	gc_values;
    u_long	gc_valuemask;
    long	window_event_mask;
    XColor	col;

    window = XCreateSimpleWindow(display, RootWindow(display, screen),
				 WIN_X, WIN_Y, WIN_W, WIN_H, 2, pen_fg, pen_bg);

    size_hints.flags = PSize | PMinSize | PMaxSize;
    size_hints.width = size_hints.min_width = size_hints.max_width = WIN_W;
    size_hints.height = size_hints.min_height = size_hints.max_height = WIN_H;

    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = progname;
    class_hints.res_class = "Sand";

    XmbSetWMProperties(display, window, "Sand", "Sand",
		       argv, argc, &size_hints, &wm_hints, &class_hints);

    window_event_mask = ButtonPressMask | KeyPressMask | StructureNotifyMask;
    XSelectInput(display, window, window_event_mask);

    gc_values.graphics_exposures = False;
    gc_values.background = pen_bg;
    gc_valuemask = GCGraphicsExposures | GCForeground | GCBackground;

    if(!XParseColor(display, cmap, "red", &col))
    {
	printf("\"red\" is not in the color database\n");
	exit(1);
    }
    if(!XAllocColor(display, cmap, &col))
    {
	printf("could not allocate color\n");
	exit(1);
    }
    gc_values.foreground = col.pixel;
    gcs[0] = XCreateGC(display, window, gc_valuemask, &gc_values);

    if(!XParseColor(display, cmap, "blue", &col))
    {
	printf("\"blue\" is not in the color database\n");
	exit(1);
    }
    if(!XAllocColor(display, cmap, &col))
    {
	printf("could not allocate color\n");
	exit(1);
    }
    gc_values.foreground = col.pixel;
    gcs[6] = XCreateGC(display, window, gc_valuemask, &gc_values);

    col.red=65535; col.blue=21845; col.green=0;
    if(!XAllocColor(display, cmap, &col))
    {
	printf("could not allocate color\n");
	exit(1);
    }
    gc_values.foreground = col.pixel;
    gcs[1] = XCreateGC(display, window, gc_valuemask, &gc_values);

    col.red=65535; col.blue=21845*2; col.green=0;
    if(!XAllocColor(display, cmap, &col))
    {
	printf("could not allocate color\n");
	exit(1);
    }
    gc_values.foreground = col.pixel;
    gcs[2] = XCreateGC(display, window, gc_valuemask, &gc_values);

    col.red=65535; col.blue=65535; col.green=0;
    if(!XAllocColor(display, cmap, &col))
    {
	printf("could not allocate color\n");
	exit(1);
    }
    gc_values.foreground = col.pixel;
    gcs[3] = XCreateGC(display, window, gc_valuemask, &gc_values);

    col.red=21845*2; col.blue=65535; col.green=0;
    if(!XAllocColor(display, cmap, &col))
    {
	printf("could not allocate color\n");
	exit(1);
    }
    gc_values.foreground = col.pixel;
    gcs[4] = XCreateGC(display, window, gc_valuemask, &gc_values);

    col.red=21845; col.blue=65535; col.green=0;
    if(!XAllocColor(display, cmap, &col))
    {
	printf("could not allocate color\n");
	exit(1);
    }
    gc_values.foreground = col.pixel;
    gcs[5] = XCreateGC(display, window, gc_valuemask, &gc_values);

    gc_values.foreground = pen_bg;
    gcblack = XCreateGC(display, window, gc_valuemask, &gc_values);

}

void CloseAll(int val)
{
    XFreeGC(display, gcs[0]);
    XFreeGC(display, gcs[1]);
    XFreeGC(display, gcs[2]);
    XFreeGC(display, gcs[3]);
    XFreeGC(display, gcs[4]);
    XFreeGC(display, gcs[5]);
    XFreeGC(display, gcs[6]);
    XFreeGC(display, gcblack);
    XCloseDisplay(display);
    exit(val);
}

void ClearWindow()
{
    XFillRectangle(display,window,gcblack,WIN_X,WIN_Y,WIN_W,WIN_H);
}    

void Delay(unsigned long value)
{
    struct timeval delay;

    delay.tv_sec  = value / 1000000;
    delay.tv_usec = value % 1000000;

    if(select(0, NULL, NULL, NULL, &delay))
	fprintf(stderr,"%s: in function Delay: select failed!\n", progname);
}       

void initdots()
{
    int k;

    pos = pos1;
    opos = pos2;

    for(k=0; k<WIN_H*WIN_W; k++)
    {
	pos1[k].gc = gcblack;
	pos2[k].gc = gcblack;
    }

    for(k=0; k<DOTS; k++)
    {
	int i,x,y;

	x = random() % WIN_W;
	y = random() % (WIN_H/3);
	
	i = y*WIN_W+x;
	dots[k].dire = 1;
	dots[k].dirw = 1;
	dots[k].dirn = 0;
	dots[k].dirs = 2;
	pos[i].vweight=(2<<5);
	pos[i].hweight=0;
	pos[i].gc = gcs[6];
	dots[k].x=x;
	dots[k].y=y;
	dots[k].sand = pos[i];

	x = random() % WIN_W;
	y = random() % (WIN_H/3) + (2*WIN_H)/3;
	
	i = y*WIN_W+x;
	dots[k+DOTS].dire = 1;
	dots[k+DOTS].dirw = 1;
	dots[k+DOTS].dirn = 2;
	dots[k+DOTS].dirs = 0;
	pos[i].vweight=-(2<<5);
	pos[i].hweight=0;
	pos[i].gc = gcs[0];
	dots[k+DOTS].x=x;
	dots[k+DOTS].y=y;
	dots[k+DOTS].sand = pos[i];
    }
}

int main(int argc, char *argv[])
{
    unsigned long plane_mask[3];
    unsigned long color[3];
    int delcount=0;

    srandom(time(NULL));
    progname = argv[0];
    display = XOpenDisplay(NULL);
    screen = DefaultScreen(display);
    cmap = DefaultColormap(display, screen);

    pen_fg = WhitePixel(display,screen);
    pen_bg = BlackPixel(display,screen);

    if (XAllocColorCells(display, cmap, False, plane_mask, 0, color, 3))
    {
	printf("Changable color map!\n");
    }
    else
    {
	printf("No changable color map!\n");
    }

    signal(SIGINT, CloseAll);
    signal(SIGTERM, CloseAll);

    InitWindow(argc, argv);

    initdots();

    XFlush(display);
    XMapWindow(display,window);

    while(1)
    {
	if(XPending(display))
	{
	    XNextEvent(display,&event);
	    switch(event.type)
	    {
		case ButtonPress:
		    if(event.xbutton.button == Button3)
		    {
			CloseAll(0);
		    }
		    else if(event.xbutton.button == Button1)
		    {
			ClearWindow();
		    }
		    break;
		case KeyPress:
		    switch(event.xkey.keycode)
		    {
			default:
			    printf("hit key %d\n",event.xkey.keycode);
			    break;
		    }
		    break;
		default:
		    break;
	    }
	}
	else	/* no event */
	{
	    int i;
	    DOT *d;

	    XDrawPoints(display, window, gcblack, points, delcount, CoordModeOrigin);

	    for(d=&dots[0],i=0; i<DOTS; d++,i++)
		XDrawPoint(display, window, d->sand.gc, d->x, d->y);

	    for(d=&dots[DOTS],i=0; i<DOTS; d++,i++)
		XDrawPoint(display, window, d->sand.gc, d->x, d->y);


	    delcount=0;
	    for(d=&dots[0],i=0; i<DOTS*2; i++)
	    {
		int x,y,r;
		
		r=random();
		x=d->x;
		y=d->y;
		if((r&3) < d->dirw) x--;
		r>>=4;
		if((r&3) < d->dire) x++;
		r>>=4;
		if((r&3) < d->dirn) y--;
		r>>=4;
		if((r&3) < d->dirs) y++;

		if(x >= WIN_W) x -= WIN_W;
		if(y >= WIN_H) y -= WIN_H;
		if(x < 0) x += WIN_W;
		if(y < 0) y += WIN_H;


		if(x != d->x || y != d->y)
		{
		    int t = x+y*WIN_W;
		    int u = (d->x)+((d->y)*WIN_W);
		    SAND *p= &pos[t];

		    if(p->gc != gcblack)
		    {
			if(p->vweight < 0) /* moving up */
			{
			    if(d->dirn != 3)
				d->dirn++;
			    if(d->dirs != 0)
				d->dirs--;
			    if(p->vweight < -64) /* moving up fast */
			    {
				if(d->dirn != 3)
				    d->dirn++;
				if(d->dirs != 0)
				    d->dirs--;
			    }
			}
			else /*moving down */
			{
			    if(d->dirs != 3)
				d->dirs++;
			    if(d->dirn != 0)
				d->dirn--;
			    if(p->vweight > 64) /* moving down fast */
			    {
				if(d->dirs != 3)
				    d->dirs++;
				if(d->dirn != 0)
				    d->dirn--;
			    }
			}

			p->vweight = (d->dirs<<5)-(d->dirn<<5);

			d->sand.gc=gcs[p->vweight/32+3];

			pos[u].gc = d->sand.gc;
			pos[u].vweight = p->vweight;
		    }
		    else
		    {
			points[delcount].x = d->x;
			points[delcount].y = d->y;
			delcount++;

			pos[t] = pos[u];
			pos[u].gc = gcblack;
			d->x=x;
			d->y=y;
		    }
		}
		d++;
	    }
	    Delay(1000);
	}
    }
    exit(0);
}
