#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
extern int errno;

void InitWindow(int argc, char *argv[]);

#define DOTS 10000
#define COLS   512
#define ROWS   256

#define PUTPIXEL(p,t,pix) if (bperpix == 32) { \
			    if (border == MSBFirst) { \
				p[t++] = (pix>>24) & 0xff; \
				p[t++] = (pix>>16) & 0xff; \
				p[t++] = (pix>>8)  & 0xff; \
				p[t++] =  pix      & 0xff; \
			    } else {  /* LSBFirst */ \
				p[t++] =  pix      & 0xff; \
				p[t++] = (pix>>8)  & 0xff; \
				p[t++] = (pix>>16) & 0xff; \
				p[t++] = (pix>>24) & 0xff; \
			    } \
			} else if (bperpix == 24) { \
			    if (border == MSBFirst) { \
				p[t++] = (pix>>16) & 0xff; \
				p[t++] = (pix>>8)  & 0xff; \
				p[t++] =  pix      & 0xff; \
			    } else {  /* LSBFirst */ \
				p[t++] =  pix      & 0xff; \
				p[t++] = (pix>>8)  & 0xff; \
				p[t++] = (pix>>16) & 0xff; \
			    } \
			} else if (bperpix == 16) { \
			    if (border == MSBFirst) { \
				p[t++] = (pix>>8)  & 0xff; \
				p[t++] =  pix      & 0xff; \
			    } else {  /* LSBFirst */ \
				p[t++] =  pix      & 0xff; \
				p[t++] = (pix>>8)  & 0xff; \
			    } \
			} else if (bperpix == 8) \
			    p[t++] =  pix      & 0xff

typedef struct {
    short x,y;
    u_char dirn;
    u_char dirs;
    u_char dire;
    u_char dirw;
} DOT;

int weights[ROWS*COLS];
DOT dots[DOTS*2];

char			*progname;
Display			*display;
int			screen, depth;
Visual			*visual;
Window			window;
XEvent			event;    
GC			gc;
XImage			*xim;
unsigned long		rmask, gmask, bmask, rrotate, grotate, brotate;
int			bperline, border, bperpix, rshift, gshift, bshift;
unsigned char		*mmbuf;
struct meteor_mem	*common_mem;
int			sig_cnt, readok;
XShmSegmentInfo		shminfo;
struct timeval		last_time, current_time;

void quit(int sig)
{
    if(sig)
	printf("Caught signal %d\n",sig);
    XShmDetach(display, &shminfo);
    XDestroyImage(xim);
    shmdt(shminfo.shmaddr);
    shmctl(shminfo.shmid, IPC_RMID, 0);
    printf("Quitting...\n");
    exit(0);
}

void Delay(unsigned long value)
{
    struct timeval delay;

    delay.tv_sec  = value / 1000000;
    delay.tv_usec = value % 1000000;

    if(select(0, NULL, NULL, NULL, &delay))
	fprintf(stderr,"%s: in function Delay: select failed!\n", progname);
}       

void InitWindow(int argc, char *argv[])
{
    XSizeHints		size_hints;
    XWMHints		wm_hints;
    XClassHint		class_hints;
    XGCValues		gc_values;
    unsigned long	gc_valuemask;
    long			window_event_mask;

    window = XCreateSimpleWindow(display, RootWindow(display, screen),
				 0, 0, COLS, ROWS, 2,
				 WhitePixel(display,screen), BlackPixel(display,screen));

    size_hints.flags = PSize | PMinSize | PMaxSize;
    size_hints.width = size_hints.min_width = size_hints.max_width = COLS;
    size_hints.height = size_hints.min_height = size_hints.max_height = ROWS;

    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    class_hints.res_name = progname;
    class_hints.res_class = "Sand";

    XmbSetWMProperties(display, window, "Sand", "Sand",
		       argv, argc, &size_hints, &wm_hints, &class_hints);

    window_event_mask = ExposureMask | ButtonPressMask | KeyPressMask |
	StructureNotifyMask;
    XSelectInput(display, window, window_event_mask);

    gc_values.graphics_exposures = False;
    gc_values.foreground = WhitePixel(display,screen);
    gc_values.background = BlackPixel(display,screen);
    gc_valuemask = GCGraphicsExposures | GCForeground | GCBackground;
    gc = XCreateGC(display, window, gc_valuemask, &gc_values);
}

int highbit(unsigned long ul)
{
    /* returns position of highest set bit in 'ul' as an integer (0-31),
       or -1 if none */

    int i;
    unsigned long hb;

    hb = 0x80;
    hb = hb << 24;   /* hb = 0x80000000UL */
    for (i=31; ((ul & hb) == 0) && i>=0;  i--, ul<<=1);
    return i;
}

void main(int argc, char *argv[])
{
    int ShmCompletionType;
    KeySym keysym;
    XComposeStatus compose;
    char keybuffer[11];
    u_long colors[7];
    int i;

    srandom(time(NULL));
    progname = argv[0];
    display = XOpenDisplay(NULL);
    screen = DefaultScreen(display);
    visual = DefaultVisual(display,screen);
    depth = DisplayPlanes(display,screen);

    if(XShmQueryExtension(display) == False)
    {
	fprintf(stderr,"Fatal: No MIT SHM X extension availible.\n");
	exit(1);
    }

    rmask = visual->red_mask;
    gmask = visual->green_mask;
    bmask = visual->blue_mask;
    rshift = 7 - highbit(rmask);
    gshift = 7 - highbit(gmask);
    bshift = 7 - highbit(bmask);
    
    for(i=0; i<7; i++)
    {
	int r,b;
	r = 255*i/7;
	b = 255-(255*i/7);
	if(rshift <= 0)
	    r <<= -rshift;
	else
	    r >>= rshift;
	r &= rmask;
	if(bshift <= 0)
	    b <<= -bshift;
	else
	    b >>= bshift;
	b &= bmask;
	colors[i] = r | b;
    }
    
    xim = XShmCreateImage(display, visual, depth, ZPixmap, NULL, &shminfo, COLS, ROWS);
    if(!xim)
    {
	fprintf(stderr,"Fatal: Could not Create XImage\n");
	exit(1);
    }
    shminfo.shmid = shmget(IPC_PRIVATE,
			    xim->bytes_per_line * xim->height, IPC_CREAT|0777);
    shminfo.shmaddr = xim->data = shmat (shminfo.shmid, 0, 0);
    shminfo.readOnly = False;
    if(XShmAttach (display, &shminfo) != True)
    {
	fprintf(stderr,"Fatal: Could not Attach SHM\n");
    	XDestroyImage(xim);
	shmdt(shminfo.shmaddr);
	shmctl(shminfo.shmid, IPC_RMID, 0);
	exit(1);
    }
    ShmCompletionType = XShmGetEventBase(display) + ShmCompletion;
    bperline = xim->bytes_per_line;
    bperpix  = xim->bits_per_pixel;
    border   = xim->byte_order;

    signal(SIGBUS, quit);
    signal(SIGINT, quit);
    signal(SIGHUP, quit);
    signal(SIGQUIT, quit);
    signal(SIGSEGV, quit);
    signal(SIGTERM, quit);

    if(bperpix != 8 && bperpix != 16 && bperpix != 32)
    {
	fprintf(stderr, "Fatal: No support to handle %d-bit TrueColor/DirectColor displays (only 16bpp now)\n", bperpix);
	exit(1);
    }
    
    memset(xim->data, 0, ROWS*bperline);
    
    for(i=0; i<DOTS; i++)
    {
	int x,y,pos;
	u_char *p=(u_char *)(xim->data);
again1:
	x = random() % COLS;
	y = random() % (ROWS/3);
	pos = (y*bperline)+(x*bperpix/8);
	if(p[pos])
	    goto again1;
	PUTPIXEL(p,pos,colors[6]);
	weights[x+(y*COLS)]=2<<5;
	dots[i].x=x;
	dots[i].y=y;
	dots[i].dire = 1;
	dots[i].dirw = 1;
	dots[i].dirn = 0;
	dots[i].dirs = 2;

again2:
	x = random() % COLS;
	y = random() % (ROWS/3) + 2*ROWS/3;
	pos = (y*bperline)+(x*bperpix/8);
	if(p[pos])
	    goto again2;
	PUTPIXEL(p,pos,colors[0]);
	weights[x+(y*COLS)]=-(2<<5);
	dots[i+DOTS].x=x;
	dots[i+DOTS].y=y;
	dots[i+DOTS].dire = 1;
	dots[i+DOTS].dirw = 1;
	dots[i+DOTS].dirn = 2;
	dots[i+DOTS].dirs = 0;
    }


    InitWindow(argc, argv);

    XFlush(display);
    XMapWindow(display,window);

    while(1)
    {
	if(XPending(display))
	{
	    XNextEvent(display, &event);
	    switch(event.type)
	    {
		case Expose:
		    /*  if(state | PAUSE)
			    XShmPutImage(display,window,gc, xim, event.xexpose.x, event.xexpose.y, event.xexpose.x, event.xexpose.y, event.xexpose.width, event.xexpose.height,True);
		     */
		    break;
		case KeyPress:
		    XLookupString(&event.xkey, keybuffer, 10, &keysym, &compose);
		    switch(keysym)
		    {
			/* Filter selection */
			case XK_q:
			    quit(0);
			default:
			    /* printf("hit key code %d\n",event.xkey.keycode);
			     */
			    break;
		    }
		    break;
		default:
		    break;
	    }
	}
	else
	{
	    DOT *d;
	    for(d=&dots[0],i=0; i<DOTS*2; i++, d++)
	    {
		int x,y,r;
		u_char *p=(u_char *)(xim->data);

		r = random();
		x = d->x;
		y = d->y;
		if((r&3) < d->dirw) x--;
		r>>=4;
		if((r&3) < d->dire) x++;
		r>>=4;
		if((r&3) < d->dirn) y--;
		r>>=4;
		if((r&3) < d->dirs) y++;

		if(x >= COLS) x -= COLS;
		if(y >= ROWS) y -= ROWS;
		if(x < 0) x += COLS;
		if(y < 0) y += ROWS;

		if(x != d->x || y != d->y)
		{
		    int t = (x * bperpix/8) + (y*bperline);
		    int u = ((d->x) * bperpix/8) + ((d->y) * bperline);
		    int w = x+(y*COLS);
		    int pix;


		    pix = p[t];
		    if(bperpix > 8)
			pix += p[t+1];
		    if(bperpix > 16)
			pix += p[t+2];
		    if(bperpix > 24)
			pix += p[t+3];

		    if(pix)
		    {
			/* moving up */
			if(weights[w] < 0)
			{
			    if(d->dirn != 3)
				d->dirn++;
			    if(d->dirs != 0)
				d->dirs--;
			    if(weights[w] < -64) /* moving up fast */
			    {
				if(d->dirn != 3)
				    d->dirn++;
				if(d->dirs != 0)
				    d->dirs--;
			    }
			}
			else if(weights[w]) /*moving down */
			{
			    if(d->dirs != 3)
				d->dirs++;
			    if(d->dirn != 0)
				d->dirn--;
			    if(weights[w] > 64) /* moving down fast */
			    {
				if(d->dirs != 3)
				    d->dirs++;
				if(d->dirn != 0)
				    d->dirn--;
			    }
			}

			w = weights[(d->x)+((d->y)*COLS)] = (d->dirs<<5)-(d->dirn<<5);

			pix = colors[w/32+3];
			PUTPIXEL(p,t,pix);
		    }
		    else
		    {
			weights[w] = weights[(d->x)+(d->y*COLS)];

			p[t++] = p[u];
			p[u++] = 0;
			if(bperpix > 8) {
			   p[t++] = p[u];
			   p[u++] = 0;
			}
			if(bperpix > 16) {
			   p[t++] = p[u];
			   p[u++] = 0;
			}
			if(bperpix > 24) {
			   p[t++] = p[u];
			   p[u++] = 0;
			}

			d->x = x;
			d->y = y;
		    }
		}
	    }

	    /* tell server to display image in SYSV IPC SHM segment */
	    XShmPutImage(display,window,gc,xim,0,0,0,0,COLS,ROWS,True);
	    Delay(1000);
	}
    }
}
